#ifndef _XC3S_Clock_H
#define _XC3S_Clock_H

#include "Prefix.h"
#include "EARTH_PT.h"
#include "XC3S_Register.h"
#include "OS_Thread.h"
#include "PCI_Device.h"

namespace EARTH {
namespace XC3S {
	class Register;

	class Clock {
	public:
		Clock(PCI::Device *device, XC3S::Register *register_) : mDevice(device), mRegister(register_)
		{
			_ASSERTE(mDevice);
			_ASSERTE(mRegister);

			mVersion = 0;
		}

		~Clock()
		{
		}

		status SetReset(uint data)
		{
			data |=   BIT_1(1) << BIT_WRITE_PCI_RESET | BIT_1(1) << BIT_WRITE_PCI_RESET_
				    | BIT_1(1) << BIT_WRITE_RAM_RESET | BIT_1(1) << BIT_WRITE_RAM_RESET_;

			status status = mDevice->SetReset(A_WRITE, data, false);
			return status;
		}

		void SetVersion(uint version)
		{
			mVersion = version;
		}

		status EnableRAM()
		{
			status status = EnableRAM(-128);
			return status;
		}

		status EnableRAM(int phase)
		{
			status status;

			mRegister->Write4(A_WRITE, BIT_1(1) << BIT_WRITE_PCI_RESET | BIT_1(1) << BIT_WRITE_PCI_RESET_);
			mRegister->Write4(A_WRITE, BIT_1(0) << BIT_WRITE_PCI_RESET | BIT_1(1) << BIT_WRITE_PCI_RESET_);
			status = Wait(WAIT_PCI);
			if (status) return status;

			mRegister->Write4(A_WRITE, BIT_1(1) << BIT_WRITE_RAM_RESET | BIT_1(1) << BIT_WRITE_RAM_RESET_);
			mRegister->Write4(A_WRITE, BIT_1(0) << BIT_WRITE_RAM_RESET | BIT_1(1) << BIT_WRITE_RAM_RESET_);
			status = Wait(WAIT_RAM);
			if (status) return status;

			OS::Thread::Sleep((1)*1000*1000);

			direction direction = (phase < 0) ? DIRECTION_DECREMENT : DIRECTION_INCREMENT;
			uint count = (phase < 0) ? -phase : +phase;

			for (uint i=0; i<count; i++) {
				status = Shift(direction);
				if (status) return status;
			}

			return status;
		}

		void Disable()
		{
			uint data =   BIT_1(1) << BIT_WRITE_PCI_RESET | BIT_1(1) << BIT_WRITE_PCI_RESET_
						| BIT_1(1) << BIT_WRITE_RAM_RESET | BIT_1(1) << BIT_WRITE_RAM_RESET_;

			mRegister->Write4(A_WRITE, data);
		}

	protected:
		enum {
			A_WRITE = 0x00,
			A_READ  = 0x00
		};

		enum {
			BIT_WRITE_RAM_INCDEC =  0,
			BIT_WRITE_RAM_ENABLE,

			BIT_WRITE_PCI_RESET  = 16,
			BIT_WRITE_RAM_RESET,

			BIT_WRITE_PCI_RESET_ = BIT_WRITE_PCI_RESET + 8,
			BIT_WRITE_RAM_RESET_ = BIT_WRITE_RAM_RESET + 8
		};

		enum {
			BIT_READ_PCI_LOCKED = 0,
			BIT_READ_RAM_LOCKED,
			BIT_READ_RAM_SHIFT_DONE
		};

		PCI::Device *const mDevice;
		XC3S::Register *const mRegister;
		uint mVersion;

		enum Wait {
			WAIT_PCI,
			WAIT_RAM
		};

		status Wait(Wait wait)
		{
			for (uint i=0; i<10; i++) {
				Status status;
				GetStatus(&status);

				bool locked;
				switch (wait) {
					case WAIT_PCI: locked = status.PciLocked;                  break;
					case WAIT_RAM: locked = status.RamLocked;                  break;
					default      : locked = false;            _ASSERTE(false); break;
				}

				if (locked) return PT::STATUS_OK;

				OS::Thread::Sleep((1)*1000*1000);
			}

			return PT::STATUS_DCM_LOCK_TIMEOUT_ERROR;
		}

		enum direction {
			DIRECTION_DECREMENT,
			DIRECTION_INCREMENT
		};

		status Shift(direction direction)
		{
			Status status;
			GetStatus(&status);

			bool done = status.RamShiftDone;

			uint data =   BIT_1(direction) << BIT_WRITE_RAM_INCDEC
						| BIT_1(1        ) << BIT_WRITE_RAM_ENABLE;

			mRegister->Write4(A_WRITE, data);

			for (uint i=0; i<10; i++) {
				for (uint j=0; j<1024; j++) {
					GetStatus(&status);
					if (status.RamShiftDone != done) return PT::STATUS_OK;
				}
				OS::Thread::Sleep((1)*1000*1000);
			}

			return PT::STATUS_DCM_SHIFT_TIMEOUT_ERROR;
		}

		struct Status {
			bool PciLocked,
				 RamLocked,
				 RamShiftDone;
		};

		void GetStatus(Status *status)
		{
			uint data;

			mRegister->Read4(A_READ, &data);

			_ASSERTE(status);
			status->PciLocked    = BIT_SHIFT_MASK(data, BIT_READ_PCI_LOCKED    , 1);
			status->RamLocked    = BIT_SHIFT_MASK(data, BIT_READ_RAM_LOCKED    , 1);
			status->RamShiftDone = BIT_SHIFT_MASK(data, BIT_READ_RAM_SHIFT_DONE, 1);
		}
	};
}
}

#endif
